import java.applet.Applet;
import java.awt.*;
import java.util.*;
import java.text.DecimalFormat;

public class JDemocApplet
    extends Applet
    implements Runnable {

    Thread th;

    int difftypes = 2;
    JDemocDim dim;
    byte[][] map;
    int[][] num;
    int mem;
    int numptr;
    Color colortype[];
    int updaterate = 1000;
    Vector changes;
    boolean friendly = true;

    static final int dx[] = { 1, 1, 0, -1, -1, -1, 0, 1 };
    static final int dy[] = { 0, 1, 1, 1, 0, -1, -1, -1 };

    static final DecimalFormat df = new DecimalFormat("000000");

    // Double buffering
    Image offscreen;
    int offscreenwidth;
    int offscreenheight;
    Graphics og;
    int width;
    int height;

    int gheight;
    int fact;
    int wfact;
    int mwidth;
    int mheight;
    int currmem;

    public void init() {

	// Getting parameters

	int xdim = 100;
	int ydim = 100;

	String s_xdim = getParameter("XDim");
	if (s_xdim != null) {
	    try {
		xdim = Integer.parseInt(s_xdim);
		if (xdim < 1)
		    xdim = 1;
	    } catch (NumberFormatException e) {}
	}

	String s_ydim = getParameter("YDim");
	if (s_ydim != null) {
	    try {
		ydim = Integer.parseInt(s_ydim);
		if (ydim < 1)
		    ydim = 1;
	    } catch (NumberFormatException e) {}
	}

	dim = new JDemocDim(xdim, ydim);

 	String s_difftypes = getParameter("DiffTypes");
	if (s_difftypes != null) {
	    try {
		difftypes = Integer.parseInt(s_difftypes);
		if (difftypes < 2)
		    difftypes = 2;
		if (difftypes > 10)
		    difftypes = 10;
	    } catch (NumberFormatException e) {}
	}

	changes = new Vector(updaterate);

 	String s_updaterate = getParameter("UpdateRate");
	if (s_updaterate != null) {
	    try {
		updaterate = Integer.parseInt(s_updaterate);
	    } catch (NumberFormatException e) {}
	}

 	String s_friendly = getParameter("Friendly");
	if (s_friendly != null) {
	    try {
		if (s_friendly.equalsIgnoreCase("false"))
		    friendly = false;
	    } catch (NumberFormatException e) {}
	}

	mem = 2000; // xdim -> max mem
	numptr = 0;
	num = new int[mem][difftypes];
	map = new byte[dim.getWidth()][dim.getHeight()];
	for (int y = 0; y < dim.getHeight(); y++)
	    for (int x = 0; x < dim.getWidth(); x++) {
		map[x][y] = (byte) (difftypes * Math.random());
		num[numptr][map[x][y]]++;
	    }
	colortype = new Color[10];
	colortype[0] = Color.blue;
	colortype[1] = Color.red;
	colortype[2] = Color.green;
	colortype[3] = Color.yellow;
	colortype[4] = Color.cyan;
	colortype[5] = Color.orange;
	colortype[6] = Color.pink;
	colortype[7] = Color.magenta;
	colortype[8] = Color.lightGray;
	colortype[9] = Color.darkGray;
    }
    
    public void start() {
	th = new Thread(this);
	th.start();
    }


    public void stop() {
	th = null;
    }

    public void run() {
	Thread me = Thread.currentThread();
	while (th == me) {
	    repaint();
	    Toolkit.getDefaultToolkit().sync();
	    Thread.yield();
	    for(int j = 0; j < updaterate; j++)
		evolve();
	    int wptr = numptr;
	    if (++numptr >= mem)
		numptr -= mem;
	    System.arraycopy(num[wptr], 0, num[numptr], 0, difftypes);
	}
    }

    public void update(Graphics g) {

	width = getSize().width;
	height = getSize().height;

	if(offscreen == null || 
	   offscreenwidth != width ||
	   offscreenheight != height) {
	    paint(g);
	}

	offScreenUpdate(og);

	g.drawImage(offscreen, 0, 0, null);

    }

    public void offScreenUpdate(Graphics g) {
	for(Enumeration e = changes.elements(); e.hasMoreElements(); ) {
	    JDemocPoint p = (JDemocPoint) e.nextElement();
	    int x = p.getX();
	    int y = p.getY();
	    g.setColor(colortype[map[x][y]]);
	    g.fillRect(x * fact, y * fact, wfact, wfact);
	}
	changes.removeAllElements();

	paintGraph(g);

	paintStatistics(g);
    }

    public void paint(Graphics g) {
	
	width = getSize().width;
	height = getSize().height;

	if(offscreen == null || 
	   offscreenwidth != width ||
	   offscreenheight != height) {

	    offscreen = createImage(width, height);
	    offscreenwidth = width;
	    offscreenheight = height;
	    og = offscreen.getGraphics();
	    og.setFont(g.getFont());

	    fact = Math.min((int) ((width - g.getFontMetrics().stringWidth("0000000000")) / dim.getWidth()),
			    (int) ((height * 0.9 - 2) /
				    dim.getHeight()));
	    if (fact < 1)
		fact = 1;
	    wfact = fact - 1;
	    if (wfact < 1)
		wfact = 1;
	    
	    mwidth = fact * dim.getWidth();
	    mheight = fact * dim.getHeight();
	    currmem = mwidth;
	    gheight = height - mheight;

	}

	og.setColor(Color.black);
	og.fillRect(0, 0, width, height);

	offScreenPaint(og);

	g.drawImage(offscreen, 0, 0, null);

    }
    
    public void offScreenPaint(Graphics g) {

	// LAYOUT...
	g.setColor(Color.white);
	g.drawLine(0, height - gheight, mwidth, height - gheight);
	g.drawLine(mwidth, 0, mwidth, height);

	for (int y = 0; y < dim.getHeight(); y++)
	    for (int x = 0; x < dim.getWidth(); x++) {
		g.setColor(colortype[map[x][y]]);
		g.fillRect(x * fact, y * fact, wfact, wfact);
	    }

	paintGraph(g);

	paintStatistics(g);
    }	

    public void paintGraph(Graphics g) {
	g.setColor(Color.black);
	g.fillRect(0, height - gheight, mwidth, gheight);
	int ptr;
	int x, y;
	for (int type = 0; type < difftypes; type++) {
	    g.setColor(colortype[type]);
	    for (int t = 0; t < currmem; t++) {
		ptr = numptr - t;
		while (ptr < 0) // just an if ?
		    ptr += mem;
		x = mwidth - 1 - t;
		y = height -
		    (int) (gheight * 
			   (((double) num[ptr][type]) / 
			    dim.getArea()));
		g.drawLine(x, y, x, y);
	    }	
	}	
    }

    public void paintStatistics(Graphics g) {
	FontMetrics fm = g.getFontMetrics();
	int fh = fm.getHeight() + 1;
	g.setColor(Color.white);
	g.drawString("Statistics:", mwidth + 3, fh);
	for (int type = 0; type < difftypes; type++) {
	    g.setColor(Color.black);
	    g.fillRect(mwidth + 3, fh * (1 + type), width - mwidth - 2, fh);
	    if (num[numptr][type] > 0) {
		g.setColor(colortype[type]);
		g.drawString(df.format(num[numptr][type]), mwidth + 3, fh * (2 + type));
	    }
	}
    }

    public void evolve() {
	JDemocPoint p =
	    new JDemocPoint((int) (dim.getWidth() * Math.random()),
			    (int) (dim.getHeight() * Math.random()));
	int d = (int) (8 * Math.random());
	JDemocPoint np = new JDemocPoint (p.getX() + dx[d],
					  p.getY() + dy[d]);

	np.fitInto(dim);
	byte oldmap = map[p.getX()][p.getY()];
	byte newmap = map[np.getX()][np.getY()];
	if (!friendly)
	    if (++newmap >= difftypes)
		newmap -= difftypes;;
	if (oldmap != newmap) {
	    num[numptr][oldmap]--;
	    num[numptr][newmap]++;
	    map[p.getX()][p.getY()] = newmap;
	    changes.addElement(p);
	}
    }

    public String getAppletInfo() {
        return "Title: JDemoc \nAuthor: Danilo Poccia, 1998 \nA simulation of a democratic environment.";
    }
  
    public String[][] getParameterInfo() {
        String[][] info = {
            {"XDim", "number", "The number of different people on the X axis."},
            {"YDim", "number", "The number of different people on the Y axis."},
            {"DiffTypes", "number", "The number of different parties to choose from."},
            {"UpdateRate", "number", "The number of people to change before updating the display."},
            {"Friendly", "boolean", "If people agree with the neighbour party or not."}
        };
        return info;
    }

}
