import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class JDemocCanvas
    extends Canvas
    implements Runnable {

    int numtypes;
    JDemocDim dim;
    byte[][] map;
    int[][] num;
    int mem;
    int numptr;
    Color colortype[];
    int updaterate;
    Vector changes;

    static final int dx[] = { 1, 1, 0, -1, -1, -1, 0, 1 };
    static final int dy[] = { 0, 1, 1, 1, 0, -1, -1, -1 };

    // DOUBLE BUFFERING
    Image offscreen;
    int offscreenwidth;
    int offscreenheight;
    Graphics og;
    int width;
    int height;

    int gheight;
    int fact;
    int wfact;
    int mwidth;
    int mheight;
    int currmem;

    public JDemocCanvas(Hashtable param) {
	super();
	dim = (JDemocDim) param.get("dim");
	numtypes = ((Integer) param.get("num")).intValue();
	updaterate = ((Integer) param.get("upd")).intValue();
	changes = new Vector(updaterate);
    }
   
    public void init() {
	mem = 1000; // 1000-> max mem
	numptr = 0;
	num = new int[mem][numtypes];
	map = new byte[dim.getWidth()][dim.getHeight()];
	for (int y = 0; y < dim.getHeight(); y++)
	    for (int x = 0; x < dim.getWidth(); x++) {
		map[x][y] = (byte) (numtypes * Math.random());
		num[numptr][map[x][y]]++;
	    }
	for (int j = 0; j < mem; j++)
	    // colortype = new Color[numtypes];
	    colortype = new Color[4]; // 4 -> max numtypes
	colortype[0] = Color.blue;
	colortype[1] = Color.red;
	colortype[2] = Color.green;
	colortype[3] = Color.yellow;
	// colortype[4] = Color.purple;
    }
    
    public void start() {
	Thread t = new Thread(this);
	t.run();
    }

    public void run() {
	for(;;) {
	    repaint();
	    Toolkit.getDefaultToolkit().sync();
	    Thread.yield();
	    for(int j = 0; j < updaterate; j++)
		evolve();
	    int wptr = numptr;
	    if (++numptr >= mem)
		numptr -= mem;
	    System.arraycopy(num[wptr], 0, num[numptr], 0, numtypes);
	}
    }

    public void update(Graphics g) {

	width = getSize().width;
	height = getSize().height;

	if(offscreen == null || 
	   offscreenwidth != width ||
	   offscreenheight != height) {
	    paint(g);
	}

	offScreenUpdate(og);

	g.drawImage(offscreen, 0, 0, null);

    }

    public void offScreenUpdate(Graphics g) {
	for(Enumeration e = changes.elements(); e.hasMoreElements(); ) {
	    JDemocPoint p = (JDemocPoint) e.nextElement();
	    int x = p.getX();
	    int y = p.getY();
	    g.setColor(colortype[map[x][y]]);
	    g.fillRect(x * fact, y * fact, wfact, wfact);
	}
	changes.removeAllElements();

	g.copyArea(0 + 1, height - gheight + 1, mwidth - 1, // - 1
		   gheight - 2, -1, 0);
	g.setColor(Color.black);
	g.drawLine(mwidth - 1, height - gheight + 1, mwidth - 1, height - 2);
	
	for (int type = 0; type < numtypes; type++) {
	    int x, y;

	    g.setColor(colortype[type]);

	    x = mwidth - 1;
	    y = height -
		(int) (gheight * 
		       (((double) num[numptr][type]) / 
			dim.getArea()));
	    g.drawLine(x, y, x, y);
	}
    }

    public void paint(Graphics g) {
	
	width = getSize().width;
	height = getSize().height;

	if(offscreen == null || 
	   offscreenwidth != width ||
	   offscreenheight != height) {

	    offscreen = createImage(width, height);
	    offscreenwidth = width;
	    offscreenheight = height;
	    og = offscreen.getGraphics();
	    og.setFont(g.getFont());

	    fact = Math.min((int) (width / dim.getWidth()),
			    (int) (((height * 0.9 - 2) /
				    dim.getHeight())));
	    if (fact < 1)
		fact = 1;
	    wfact = fact - 1;
	    if (wfact < 1)
		wfact = 1;
	    
	    mwidth = fact * dim.getWidth();
	    mheight = fact * dim.getHeight();
	    currmem = mwidth;
	    gheight = height - mheight;

	}

	og.setColor(Color.black);
	og.fillRect(0, 0, width, height);

	offScreenPaint(og);

	g.drawImage(offscreen, 0, 0, null);

    }
    
    public void offScreenPaint(Graphics g) {

	// LAYOUT...
	g.setColor(Color.white);
	g.drawLine(0, height - gheight, mwidth, height - gheight);
	g.drawLine(mwidth, 0, mwidth, height);

	for (int y = 0; y < dim.getHeight(); y++)
	    for (int x = 0; x < dim.getWidth(); x++) {
		g.setColor(colortype[map[x][y]]);
		g.fillRect(x * fact, y * fact, wfact, wfact);
	    }
	int ptr;
	int x, y;
	for (int type = 0; type < numtypes; type++) {
	    g.setColor(colortype[type]);
	    for (int t = 0; t < currmem; t++) {
		ptr = numptr - t;
		while (ptr < 0) // just an if ?
		    ptr += mem;
		x = mwidth - 1 - t;
		y = height -
		    (int) (gheight * 
			   (((double) num[ptr][type]) / 
			    dim.getArea()));
		g.drawLine(x, y, x, y);
	    }	
	}	
    }	

    public void evolve() {
	JDemocPoint p =
	    new JDemocPoint((int) (dim.getWidth() * Math.random()),
			    (int) (dim.getHeight() * Math.random()));
	int d = (int) (8 * Math.random());
	JDemocPoint np = new JDemocPoint (p.getX() + dx[d],
					  p.getY() + dy[d]);
	np.fitInto(dim);
	byte oldmap = map[p.getX()][p.getY()];
	byte newmap = map[np.getX()][np.getY()];
	if (oldmap != newmap) {
	    num[numptr][oldmap]--;
	    num[numptr][newmap]++;
	    map[p.getX()][p.getY()] = newmap;
	    changes.addElement(p);
	}
    }

}
